/* Copyright (C) Burr Settles, Department of Computer Sciences, UW-Madison.
This file is part of "AMIL" (Active Machine Learning Library).
http://pages.cs.wisc.edu/~bsettles/amil
This software is licensed under the terms of the GPL. For more information, 
see "LICENSE.html" included with this distribution. */

package amil.classify;

import amil.classify.*;
import amil.types.*;
import amil.util.*;
import java.util.*;

/**
 * Multiple Instance Logistic Regression model (Ray & Craven, 2005).
 *
 * @author Burr Settles &lt;&gt;
 * @version $Rev$
 */
public class MILogisticRegression extends Model {
/*	public final boolean NORMALIZE_BAGS = true;*/
	private final boolean CLAMP_QUERIES = false;
	
	private double
		BFGS_EPS = 1e-6,	// magnitude of hessian / 1e-6
		BFGS_TOLX = 1e-4;	// x tolerance / 1e-4
	private int
		BFGS_STORE = 10,	// number of updates to store / 30
		BFGS_ITER = 100;	// max number of iterations / 150
	
	Random rand;
	
    public MILogisticRegression(CombFn cf, Random r) {
		this.combiner = cf;
		this.rand = r;
    }
	public MILogisticRegression(CombFn cf, int seed) {
		this(cf, new Random(seed));
	}
	public MILogisticRegression(int seed) {
		this(new SoftMax(), new Random(seed));
	}
	public MILogisticRegression() {
		this(new SoftMax(), new Random());
	}
	
	/**
	 * Initialize the model randomly.
	 */
	public void initialize(BagList bags) {
		parameters = new double[bags.fvLength()+1];
		for (int i=0; i<parameters.length; i++) {
/*			parameters[i] = rand.nextGaussian();*/
			parameters[i] = rand.nextDouble();
		}
	}
	
	/**
	 * Learn a new model from training set "bags." Returns the final objective function value on train data.
	 */
	public double train(BagList bags) {
		// initialize parameter space if it's null or doesn't match the data
		if (parameters == null || parameters.length != bags.fvLength()+1)
			initialize(bags);			

/*		System.out.println("#### PARAMS("+parameters.length+"):"+MiscMath.dumpArray(parameters));*/

		double[] diag = new double[parameters.length];
		int[] iprint = new int[]{-1,0};
		int[] iflag = new int[]{0};
/*		boolean ex = false;*/

		for (int iter = 0; iter<BFGS_ITER && (iter==0 || iflag[0]!=0); ++iter) {
			// Compute SSE and gradient of SSE w.r.t. weights.
			
			double value = valueAt(bags);
			double[] gradient = gradientAt(bags);

			// Now fudge the error and gradient so that we get MSE and
			// the gradient is gradient of MSE (not SSE) w.r.t weights.
/*			value /= parameters.length;
			for (int k=0; k<gradient.length; k++)
				gradient[k] /= parameters.length;
*/
			try {
				LBFGS.lbfgs(parameters.length, BFGS_STORE, parameters, value, gradient, false, diag, iprint, BFGS_EPS, BFGS_TOLX, iflag);
			}
			catch (LBFGS.ExceptionWithIflag e) {
/*				ex = true;				*/
/*				continue;*/
			}
		}
		return valueAt(bags);
	}
	
	/**
	 * Take a list of test bags and return an array of prediction values per bag.
	 */
	public double[] classifyBags(BagList bags, double[] params) {
		double[] ret = new double[bags.size()];
		double[][] insts = classifyInstances(bags, params);
		for (int i=0; i<insts.length; i++) {
			ret[i] = combiner.combine(insts[i]);
		}
		return ret; 
	}
	
	/**
	 * Take a list of test bags and return an array of prediction values for constituent instances.
	 */
	public double[][] classifyInstances(Instance[][] instances, double[] params) {
		double[][] ret = new double[instances.length][];
		for (int i=0; i<instances.length; i++) {
			ret[i] = new double[instances[i].length];
			for (int j=0; j<instances[i].length; j++) {
				double[] point = instances[i][j].getData();
				double sum = params[params.length-1]; //bias term
				for (int k=0; k<point.length; k++) {
					sum += point[k] * params[k];
				}
				ret[i][j] = 1.0/(1.0+Math.exp(-sum));
			}
		}
		return ret;
	}
	
	/**
	 * Compute the gradient of the objective function w.r.t. model parameters given the list of bags.
	 */
	public double[] gradientAt(BagList bags, double[] params, double[] scores) {
		double[] ret = new double[params.length];
		Arrays.fill(ret,0.0);
//		double[] scores = classifyBags(bags, params);
		double[][] instScores = classifyInstances(bags, params);
		Instance[][] ilist = bags.toInstanceArray();
		
		int skipped=0;
		int notskipped=0;
		
		for (int i=0; i<scores.length; i++) {
			Bag b = bags.get(i);	
			// first, skip this bag if we're getting it "too" correct...
			if (b.isPositive() && scores[i] > b.getTarget() ||
				!b.isPositive() && scores[i] < b.getTarget()) {
				skipped++;
				continue;
			}
			notskipped++;
			if (CLAMP_QUERIES && bags.get(i).isSourceBag()) {
				for (int j=0; j<ilist[i].length; j++)
					if (ilist[i][j].isQueried() && !ilist[i][j].isPositive())
//					if (ilist[i][j].isQueried())
						instScores[i][j] = ilist[i][j].getTarget();
				scores[i] = combiner.combine(instScores[i]);
			}
			double[] combGradient = combiner.gradientAt(instScores[i]);
			double[] sums = new double[params.length];
			Arrays.fill(sums,0.0);
			for (int j=0; j<instScores[i].length; j++) {
				double[] point = ilist[i][j].getData();
				// features first...
				for (int k=0; k<params.length-1; k++) {
					sums[k] += combGradient[j] * instScores[i][j]*(1-instScores[i][j]) * point[k];
				}
				// then the bias parameter "b"
				sums[params.length-1] += combGradient[j] * instScores[i][j]*(1-instScores[i][j]);
			}
			// done. construct return vector
			double term1 = -2 * (bags.get(i).getTarget() - scores[i]); // outermost term (deriv. bag error)
			for (int k=0; k<sums.length; k++) {
				if (weightBags())
					ret[k] += bags.get(i).getWeight() * term1 * sums[k];
				else
					ret[k] += term1 * sums[k];
			}
		}
		
//		System.out.println("#### BAGS: "+bags.size()+" ("+bags.numPos()+"+, "+bags.numNeg()+"-) / SKIPPED: "+skipped+" ("+Math.round((100.0*skipped)/bags.size())+"%)");
/*		if (skipped > 0)
			System.out.print(skipped+"|");
*/			

/*		if (skipped > 0)
			System.out.print("["+bags.size()+":"+skipped+":"+notskipped+"]");
*/	
		
		return ret;
	}
	
	/**
	 * Returns the prediction scores for bags (legacy).
	 */
	public double[] scoresAt(BagList bags, double[] params) {
		return classifyBags(bags, params);
	}
	
	/**
	 * Calculates the objective function value of the model parameters on the bag list.
	 */
	public double valueAt(BagList bags, double[] params) {
		double objfn = 0;
		double[] scores = scoresAt(bags, params);
//		double[] targets = bags.toTargetArray();
/*		boolean[] pos = bags.isPositiveArray();*/
		for (int i=0; i<scores.length; i++) {
			if (weightBags())
				objfn += bags.get(i).getWeight() * Math.pow(bags.get(i).getTarget() - scores[i], 2);
			else
				objfn += Math.pow(bags.get(i).getTarget() - scores[i], 2);
		}			
		return objfn;
	}

	/**
	 * Return model type as string.
	 */
	public String toString() {
		return "MILR";
	}
}
