/* Copyright (C) Burr Settles, Department of Computer Sciences, UW-Madison.
This file is part of "AMIL" (Active Machine Learning Library).
http://pages.cs.wisc.edu/~bsettles/amil
This software is licensed under the terms of the GPL. For more information, 
see "LICENSE.html" included with this distribution. */

package amil.classify;

/**
 * SoftMax combining function.
 *
 * @author Burr Settles &lt;&gt;
 * @version $Rev$
 */
public class SoftMax extends CombFn {
	private double alpha; //parameter controlling "max" vs. "average" behavior

	/**
	 * The default softmax combiner uses <i>alpha</i> = 2.5.
	 */
	public SoftMax() { this(2.5); }
	public SoftMax(double alpha) { this.alpha = alpha;}	

	/**
	 * Combine values in <i>d</i> using the softmax function.
	 */
	public double combine(double[] d){
		return (numer(d)/denom(d));
	}

	private double numer(double[] d){
		double numer = 0;
		for (int i=0; i<d.length; i++){
			numer += d[i] * Math.exp(alpha*d[i]);
		}
		return numer;
	}

	private double denom(double[] d){
		double denom = 0;
		for (int i=0; i<d.length; i++){
			denom += Math.exp(alpha*d[i]);
		}
		return denom;
	}

	/**
	 * Calculate the gradient of softmax output wrt each input value in <i>d</i>.
	 */
	public double[] gradientAt(double[] d) {
		double[] ret = new double[d.length];
		double val = combine(d);
		double den = denom(d);
		double term3 = alpha*val;
		for (int i=0; i<d.length; i++){
			double term1 = Math.exp(alpha*d[i])/den;
			double term2 = alpha*d[i];
			ret[i] = term1*(1+term2-term3);
		}
		return ret;
	}

	/**
	 * Return a string representation.
	 */
	public String toString(){
		return "SoftMax("+alpha+")";
	}
}
