/* Copyright (C) Burr Settles, Department of Computer Sciences, UW-Madison.
This file is part of "AMIL" (Active Machine Learning Library).
http://pages.cs.wisc.edu/~bsettles/amil
This software is licensed under the terms of the GPL. For more information, 
see "LICENSE.html" included with this distribution. */

package amil.classify.active;

import amil.classify.*;
import amil.types.*;
import amil.util.*;
import java.util.*;

/**
 * <p>Expected Gradient Length (EGL) instance-query strategy (Settles et. al, 2008). Query the instance that generates the largest expected change in the learning gradient after adding this instance. 
	<p>Note that each partial derivative in the training gradient is obtained by differentiating the objective function (squared loss) using the chain rule. If training has converged properly, these should all be nearly zero. So we simplify things here by only calculating the expected gradient for each candidate instance (pseudobag), since it will make the only non-negligible contribution to the overall gradient.</p>
 *
 * @author Burr Settles &lt;&gt;
 * @version $Rev$
 */
public final class InstanceEGL extends QuerySelector {
	public BagList select(Model model, BagList bags, int num) {
		BagList ret = new BagList();
		Instance[][] ilist = bags.toInstanceArray();
		double[][] preds = model.classifyInstances(bags);
		ArrayList v = new ArrayList();
		
		for (int i=0; i<ilist.length; i++) {
			// only query from pos+ bags
			if (bags.get(i).isPositive()) {
				for (int j=0; j<ilist[i].length; j++) {
					// only unqueried instances
					if (!ilist[i][j].isQueried()) {
						double[] pos_vect = gradientVector(model, preds[i][j], ilist[i][j].getData(), 1.0);
						double[] neg_vect = gradientVector(model, preds[i][j], ilist[i][j].getData(), 0.0);
						double score = preds[i][j] * gradientSize(pos_vect) + 
							(1-preds[i][j]) * gradientSize(neg_vect);
						// be sure to weight the gradient appropriately if we're normalizing bags+pseudobags
						if (model.weightBags())
							score /= (bags.get(i).getPseudoCount() + 1);
						v.add(new Ranker(score, ilist[i][j]));
					}
				}				
			}
		}


		Collections.sort(v);

		for (int n=0; n<num; n++) {
			Ranker r = (Ranker)v.get(n);
			ret.add(new PseudoBag((Instance)r.getObject()));
			System.out.println("\t"+r);
		}
		return ret; 
	}
	
	// this computes the size of the vector...
	private double gradientSize(double[] v) {
		double sum = 0;
		for (int i=0; i<v.length; i++) {
			sum += Math.pow(v[i], 2);
		}
		return Math.sqrt(sum);
	}
	
	
	private double[] gradientVector(Model model, double score, double[] points, double target) {
		double[] ret = new double[model.getParams().length];
		double term1 = -2 * (target - score); //outermost term
		//first the bias term
		ret[ret.length-1] = term1 * score * (1-score);
		//now the rest of the features
		for (int k=0; k<points.length; k++) {
			ret[k] = term1 * score * (1-score) * points[k];
		}
		return ret;	
	}
	
}
