/* Copyright (C) Burr Settles, Department of Computer Sciences, UW-Madison.
This file is part of "AMIL" (Active Machine Learning Library).
http://pages.cs.wisc.edu/~bsettles/amil
This software is licensed under the terms of the GPL. For more information, 
see "LICENSE.html" included with this distribution. */

package amil.types;

import amil.types.*;


/**
 * Bag class. In MI learning, negative bags contain only negative instances and positive bags contain at least one positive (but any number of negatives).
 *
 * @author Burr Settles &lt;&gt;
 * @version $Rev$
 */
public class Bag {
	protected String bagName;
	protected Instance[] instances;
	protected boolean target;
	protected boolean queried = false;
	protected double time;
	protected int pseudoCount = 1; // number of pseduobags that have come from this bag, plus itself.

    public Bag(String bn, Instance[] ilist, boolean t) {
		this.bagName = bn;
		this.instances = ilist;
		this.target = t;
		if (ilist != null)
			for (int i=0; i<ilist.length; i++)
				ilist[i].setBag(this);
    }

    public Bag(String bn, Instance[] ilist, double t) {
		this(bn, ilist, (t > 0.5));
    }

    public Bag(String bn, Instance[] ilist) {
		this(bn, ilist, false);
		double ct=0;
		for (int j=0; j<ilist.length; j++) {
			ct += (ilist[j].isPositive()) ? 1 : 0;
		}
		this.target = (ct > 0);
    }

    public Bag(String bn) {
		this(bn, null, 0);
    }

	/**
	 * setTime
	 */
	public void setTime(double time) {
		this.time = time;
	}
	/**
	 * getTime
	 */
	public double getTime() {
		return this.time;
	}
	
	/**
	 * getSize
	 */
	public int size() {
		return instances.length;
	}
	

	/**
	 * Return array of constituent instance objects.
	 */
	public Instance[] getInstances() {
		return instances;
	}
	
	/**
	 * Return specific instace
	 */
	public Instance getInstanceAt(int index) {
		return instances[index];
	}
	

	/**
	 * Return the target class value (0,1) for the bag.
	 */
	public double getTarget() {
/*		return (target) ? 1-1e-5 : 1e-5;*/
		return (target) ? 1.0 : 0.0;
	}
	public boolean isPositive() {
		return target;
	}	
	
	/**
	 * Return the size of feature vectors describing the instances.
	 */
	public int fvLength() {
		if (instances.length > 0)
			return (instances[0].fvLength());
		else
			return 0;
	}
	
	/**
	 * Bag ID "name."
	 */
	public String getName() {
		return bagName;
	}
	
	/**
	 * String representation.
	 */
	public String toString() {
		return bagName+"("+instances.length+")"+((target) ? "+" : "-")+" <"+time+">";
/*		return bagName+"("+instances.length+")"+((target) ? "+" : "-")+"<"+getWeight()+">";*/
	}
	
	/**
	 * Number of derivative pseudobags that have been queried, plus this bag.
	 */
	public int getPseudoCount() {
		return pseudoCount;
	}

	/**
	 * The weight this bag has on the objective function.
	 */
	public double getWeight() {
		return 1.0/pseudoCount;
	}
	
	/**
	 * Increment pseudocount (when pseudobags are made from instance queries).
	 */
	public void incPseudoCount() {
		pseudoCount++;
	}
	

	/**
	 * Has this bag been fully queried?
	 */
	public boolean isQueried() {
		return queried;
	}
	
	/**
	 * Is this a source bag (as opposed to a pseudobag)?
	 */
	public boolean isSourceBag() {
		return true;
	}

	/**
	 * Set if this bag has been fully queried.
	 */
	public void setQueried() {
		queried = true;
	}
	
	/**
	 * Reset the weight/pseudocounts and query flags for this bag and its instances.
	 */
	public void reset() {
		pseudoCount = 1;
		queried = false;
		for (int i=0; i<instances.length; i++)
			instances[i].reset();
	}
	
	
}
