/* Copyright (C) Burr Settles, Department of Computer Sciences, UW-Madison.
This file is part of "AMIL" (Active Machine Learning Library).
http://pages.cs.wisc.edu/~bsettles/amil
This software is licensed under the terms of the GPL. For more information, 
see "LICENSE.html" included with this distribution. */

package amil.types;

import java.io.*;
import java.util.*;
import amil.types.*;
import amil.util.*;

/**
 * Convenience class for maintaining Bag objects and learning MI models.
 *
 * @author Burr Settles &lt;&gt;
 * @version $Rev$
 */
public class BagList {
	ArrayList bags;
	
    public BagList() {
		bags = new ArrayList();
    }
	public BagList(int capacity) {
		bags = new ArrayList(capacity);
	}
	
	/**
	 * Add a bag
	 */
	public boolean add(Bag b) {
		return bags.add(b);
	}
	public void add(BagList blist) {
		for (int i=0; i<blist.size(); i++) {
			bags.add(blist.get(i));
		}
	}
	
	/**
	 * Return bag at index <i>i</i>.
	 */
	public Bag get(int i) {
		return (Bag)bags.get(i);
	}
	
	/**
	 * Return number of positive instances in this BagList.
	 */
	public int numPos() {
		int p=0;
		for (int i=0; i<bags.size(); i++)
			if (((Bag)bags.get(i)).isPositive())
				p++;
		return p;
	}

	/**
	 * Return number of negative instances in this BagList.
	 */
	public int numNeg() {
		int n=0;
		for (int i=0; i<bags.size(); i++)
			if (!((Bag)bags.get(i)).isPositive())
				n++;
		return n;		
	}
	
	/**
	 * getTime
	 */
	public double getTime() {
		double time = 0;
		for (int i=0; i<bags.size(); i++) {
/*			Bag b = (Bag)bags.get(i);
			System.out.println("!!!!"+b.getTime());*/
			time += ((Bag)bags.get(i)).getTime();			
		}
		return time;
	}
	
	
	/**
	 * Remove bag at index <i>i</i>.
	 */
	public Bag remove(int i) {
		return (Bag)bags.remove(i);
	}

	/**
	 * Bag size.
	 */
	public int size() {
		return bags.size();
	}

	/**
	 * Returns the bag contents as an array.
	 */
	public Bag[] toBagArray() {
		Bag[] ret = new Bag[bags.size()];
		for (int i=0; i<bags.size(); i++)
			ret[i] = (Bag)bags.get(i);
		return ret;
	}
	/**
	 * See toBagArray().
	 */
	public Bag[] toArray() {
		return toBagArray();
	}
	
	/**
	 * Returns a 2D array: first index corresponds to bags, the second the instance instide each bag.
	 */
	public Instance[][] toInstanceArray() {
		Instance[][] ret = new Instance[bags.size()][];
		for (int i=0; i<ret.length; i++) {
			ret[i] = ((Bag)bags.get(i)).getInstances();
		}
		return ret;
	}

	/**
	 * Return array of class target values (0,1) for each bag.
	 */
	public double[] toTargetArray() {
		double[] ret = new double[bags.size()];
		for (int i=0; i<bags.size(); i++)
			ret[i] = ((Bag)bags.get(i)).getTarget();
		return ret;
	}

	/**
	 * Return boolean array of class values for each bag.
	 */
	public boolean[] isPositiveArray() {
		boolean[] ret = new boolean[bags.size()];
		for (int i=0; i<bags.size(); i++)
			ret[i] = ((Bag)bags.get(i)).isPositive();
		return ret;
	}
	
	/**
	 * Return the size of feature vectors describing the instances.
	 */
	public int fvLength() {
		if (bags.size() > 0)
			return ((Bag)bags.get(0)).fvLength();
		else
			return 0;
	}
	
	/**
	 * Randomly split up baglist into array of baglists based on fold/proportion paramters
	 */
	public BagList[] split (Random r, double[] proportions) {
	    ArrayList shuffled = (ArrayList) this.bags.clone();
			Collections.shuffle (shuffled, r);
	    return splitInOrder(shuffled, proportions);
	}
	public BagList[] split (double[] proportions) {
		return split (new Random(), proportions);
	}
	public BagList[] split (Random r, int numFolds) {
		double p[] = new double[numFolds];
		for (int i=0; i<p.length; i++)
			p[i] = 1.0/numFolds;
		return split (r, p);
	}
	public BagList[] split (int numFolds) {
		return split (new Random(), numFolds);
	}
	

	/**
	 * splitInOrder
	 */
	public BagList[] splitInOrder (double[] proportions) {
    	return splitInOrder((ArrayList)this.bags.clone(), proportions);
	}
	private static BagList[] splitInOrder (List bags, double[] proportions) {
		BagList[] ret = new BagList[proportions.length];
    	double[] maxind = MiscMath.normalize(proportions);
		for (int i=0; i<maxind.length; i++) {
			ret[i] = new BagList();
			if (i > 0)
				maxind[i] += maxind[i-1];
		}
		for (int i = 0; i < maxind.length; i++)
			maxind[i] = Math.rint (maxind[i] * bags.size());
		int j=0;
		for (int i=0; i<bags.size(); i++) {
			while (i >= maxind[j])
				j++;
			ret[j].add((Bag)bags.get(i));
		}
		return ret;
	}
	
	/**
	 * Sample "pos" positives and "neg" negatives from the baglist in index [0], the remainder in [1]
	 */
	public BagList[] sample(int pos, int neg) {
		return sample(new Random(), pos, neg);
	}
	
	/**
	 * Same as above, with random seed <i>r</i>.
	 */
	public BagList[] sample(Random r, int pos, int neg) {
		ArrayList shuffled = (ArrayList) this.bags.clone();
			Collections.shuffle (shuffled, r);
		BagList[] ret = new BagList[2];
		ret[0] = new BagList();
		ret[1] = new BagList();
		int p=0,n=0;
		for (int i=0; i<shuffled.size(); i++) {
			Bag b = (Bag)shuffled.get(i);
			if (b.isPositive() && p<pos) {
				ret[0].add(b);
				p++;
			}
			else if (!b.isPositive() && n<neg) {
				ret[0].add(b);
				n++;				
			}
			else
				ret[1].add(b);
		}
		return ret;
	}
	
	/**
	 * Read a file with one bag name per line and return an array of BagLists with bags listed in the file in index [0], and the remainder in index [1]... useful for keeping consistent data partitions for cross-validation, etc.
	 */
	public BagList[] partition(String filename) {
		BagList[] ret = new BagList[2];
		ret[0] = new BagList(); // where the matching bags go
		ret[1] = new BagList();
		HashSet hash = new HashSet();
		// build a hash (bagid as key) of vectors (holding bags' instances)
        try {
	        BufferedReader br = new BufferedReader(new FileReader(filename));
	        String line;
			// store a list of all the names in the hash set
	        while ((line = br.readLine()) != null)
				hash.add(line.trim());
			// walk length of this baglist, partitioning appropriately
			for (int i=0; i<bags.size(); i++) {
				Bag b = (Bag)bags.get(i);
				// if it's a match, pluck it out
				if (hash.contains(b.getName()))
					ret[0].add(b);
				// otherwise, leave it with the rest...
				else
					ret[1].add(b);
			}
        } catch (Exception e) {
            System.err.println("Exception in BagList (partition): " + e);
        }
		return ret;
	}
	
	
	/**
	 * String representation.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (int i=0; i<bags.size(); i++) {
			sb.append("["+i+"]"+bags.get(i).toString()+"\n");
		}
		return sb.toString();
	}
	
	/**
	 * Reset the cached active learning/query info for all bags/instances within.
	 */
	public void reset() {
		for (int i=0; i<bags.size(); i++)
			((Bag)bags.get(i)).reset();
	}
	

}
