/* Copyright (C) Burr Settles, Department of Computer Sciences, UW-Madison.
This file is part of "AMIL" (Active Machine Learning Library).
http://pages.cs.wisc.edu/~bsettles/amil
This software is licensed under the terms of the GPL. For more information, 
see "LICENSE.html" included with this distribution. */

package amil.util;

import java.io.*;
import java.util.*;
import amil.types.*;

/**
 * <p>Reads a C4.5-formatted data file (".data" file only, no ".names" file needed). More specifically, the data must be a comma-delimited list of features with the instance class value at the end:</p>

<pre>bagid,instid,feat1,feat2,feat3,...,class.</pre>

<p>For example, a data set with three bags of varying sizes and three features per instance might read:</p>

<pre>
bag1,inst1a,0.00,2.51,0.20,0.
bag1,inst1b,0.10,0.03,1.28,1.
bag2,inst2a,0.00,4.44,0.21,0.
bag2,inst2b,0.01,0.98,0.32,0.
bag2,inst2c,0.00,1.35,0.67,0.
bag3,inst3a,0.25,0.11,0.99,1.
</pre>

<p>Correct bag labels are automatically generated from the instance labels... in the above example, bag1 and bag3 will be assigned positive labels. The first two elements are assumed to be the bag ID and instance ID by default, but these can be overridden. All feature values should be real numbers, and the final element (class label) must be either 0 or 1.</p>
 *
 * @author Burr Settles &lt;&gt;
 * @version $Rev$
 */
public final class DataReader {
	/**
	 * By default, the first 2 values are the bag ID and instance ID, respectively.
	 */
	public static BagList readData(String filename) {
		return readData(filename, 0, 1);
	}

	/**
	 * If the bag/instance IDs aren't the first 2 elements per line, their indexes can be specified.
	 */
	public static BagList readData(String filename, int bagid, int instid) {
		HashMap hash = new HashMap();
		BagList ret = new BagList();
		// build a hash (bagid as key) of vectors (holding bags' instances)
        try {
	        BufferedReader br = new BufferedReader(new FileReader(filename));
	        String line;
	        while ((line = br.readLine()) != null) {
				String[] feats = line.trim().split (",");
				double[] data = new double[feats.length-3];
				int j=0;
				for (int i=0; i<feats.length-1; i++) {
					if (i != bagid && i != instid) {
						data[j] = Double.parseDouble(feats[i]);
						j++;
					}
				}
				String bagName = feats[bagid];
				String instName = feats[instid];
				double target = Double.parseDouble(feats[feats.length-1]);
				Instance newInst = new Instance(instName, data, target);
				
				if (hash.containsKey(bagName)) {
					Vector v = (Vector)hash.get(bagName);
					v.add(newInst);
				} else {
					Vector v = new Vector();
					v.add(newInst);
					hash.put(bagName, v);
				}				
			}
        } catch (Exception e) {
            System.err.println("Exception in DataReader: " + e);
        }

		// pass through the hash to build the BagList (ret)
		Set keys = hash.keySet();
		Iterator kit = keys.iterator();
		while (kit.hasNext()) {
			String bagName = (String)kit.next();
			Vector v = (Vector)hash.get(bagName);
			Instance[] ilist = new Instance[v.size()];
			for (int i=0; i<ilist.length; i++) 
				ilist[i] = (Instance)v.get(i);
			ret.add(new Bag(bagName, ilist));
		}
		// done!
		return ret;
	}
}
