/* Copyright (C) Burr Settles, Department of Computer Sciences, UW-Madison.
This file is part of "AMIL" (Active Machine Learning Library).
http://pages.cs.wisc.edu/~bsettles/amil
This software is licensed under the terms of the GPL. For more information, 
see "LICENSE.html" included with this distribution. */

package amil.util;

import java.util.*;
import java.text.*;

/**
 * Collection of helpful static routines for averages, confidence intervals, uncertainty measures, etc.
 *
 * @author Burr Settles &lt;&gt;
 * @version $Rev$
 */
public class MiscMath {
    public MiscMath() {}

	/**
	 * GINI impurity measure for probability "p"
	 */
	public static double gini(double p) {
		return 1 - (Math.pow(p,2) + Math.pow(1-p,2));
	}
	/**
	 * entropy impurity measure for probability "p"
	 */
	public static double entropy(double p) {
		if (p < 1e-9 || p > 1-1e-9)
			return 0;
		else
			return - ((p)*(Math.log(p)/Math.log(2)) 
				+ (1-p)*(Math.log(1-p)/Math.log(2)));
	}

	/**
	 * minimum value of array "x"
	 */
	public static double min(double[] x) {
		double r=Double.MAX_VALUE;
		for (int i=0; i<x.length; i++)
			if (r > x[i])
				r = x[i];
		return r;
	}
	
	/**
	 * dumpArray
	 */
	public static String dumpArray(double[] x) {
		StringBuffer sb = new StringBuffer();
		DecimalFormat df = new DecimalFormat("####.######");
		for (int i=0; i<x.length; i++) {
			sb.append("["+df.format(x[i])+"]");
		}
		return sb.toString();
	}
	
	
	/**
	 * stratify so that the values of "x" all run from 0 to 1
	 */
	public static double[] stratify(double[] x) {
		double [] ret = new double[x.length];
		double min = min(x);
		double max = max(x);
		for (int i=0; i<x.length; i++) {
			ret[i] = (x[i]-min)/max;
		}
		return ret;
	}
	
	
	/**
	 * normalize values in array "x" to sum to 1
	 */
	public static double[] normalize(double[] x) {
		double [] ret = new double[x.length];
		double sum = 0;
		for (int i=0; i<x.length; i++)
			sum += x[i];
		for (int i=0; i<x.length; i++)
			ret[i] = x[i]/sum;
		return ret;
	}
	
	/**
	 * maximum value in array "x"
	 */
	public static double max(double[] x) {
		double r=Double.MIN_VALUE;
		for (int i=0; i<x.length; i++)
			if (r < x[i])
				r = x[i];
		return r;
	}

	/**
	 * sum array "x"
	 */
	public static double sum(double[] x) {
		double r = 0;
		for (int i=0; i<x.length; i++)
			r += x[i];
		return r;
	}

	/**
	 * mean value of array "x"
	 */
	static public double average(double[] x) {
		double r=0;
		for (int i=0; i<x.length; i++)
			r+=x[i];
		return (r/x.length);
	}
	
	/**
	 * standard deviation of array "x"
	 */
	static public double stdev(double[] x) {
		double mean = average(x);
		double r=0;
		for (int i=0; i<x.length; i++)
			r += Math.pow(x[i]-mean,2);
		return Math.sqrt(r/(x.length-1));
	}

	/**
	 * confidence interval for array "x"
	 */
	static public double confidence(double[] x, double coeff) {
		return coeff*(stdev(x)/Math.sqrt(x.length));
	}

	/**
	 * default 95% confidence interval
	 */
	static public double confidence(double[] x) {
		return confidence(x, 1.96);
	}
}
